/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

public class BuildMessages {
	public BuildMessages() {
		buildMessageList = new LinkedList<BuildMessage>();
	}

	public void add(BuildMessage buildMessage) {
		buildMessageList.add(buildMessage);
		switch (buildMessage.type) {
			case Error:
				errors += 1; break;
			case FatalError:
				fatals += 1; break;
			case Warning:
				warnings += 1; break;
			case Note:
				notes += 1; break;
			case Exception:
				exceptions += 1; break;
		}
	}

	public synchronized void add(BuildMessages buildMessages) {
		buildMessageList.addAll(buildMessages.buildMessageList);
		errors += buildMessages.errors;
		fatals += buildMessages.fatals;
		warnings += buildMessages.warnings;
		notes += buildMessages.notes;
		exceptions += buildMessages.exceptions;
	}

	public Collection<BuildMessage> iterable() {
		return buildMessageList;
	}

	public BuildMessage getLastBuildMessage() {
		if (buildMessageList.isEmpty()) {
			return null;
		} else {
			return buildMessageList.get(buildMessageList.size() - 1);
		}
	}

	public boolean hasErrors() {
		return getErrorsCount() != 0;
	}

	public int getErrorsCount() {
		return errors + fatals;
	}

	public int getFatalErrorsCount() {
		return fatals;
	}

	public int getWarningsCount() {
		return warnings;
	}

	public int getNotesCount() {
		return notes;
	}

	public int getExceptionsCount() {
		return exceptions;
	}

	private int errors = 0;
	private int fatals = 0;
	private int warnings = 0;
	private int notes = 0;
	private int exceptions = 0;
	private final List<BuildMessage> buildMessageList;
}
