/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build.internals;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

public class CommandLineTool {
	protected class StreamGrabber extends Thread {
		public StreamGrabber(InputStream stream, boolean isErrorStream) {
			this.isErrorStream = isErrorStream;
			this.reader = new BufferedReader(new InputStreamReader(stream));
		}

		@Override
		public void run() {
			String line = null;
			try {
				while ((line = this.reader.readLine()) != null) {
					if (this.isErrorStream) {
						error.add(line);
					} else {
						output.add(line);
					}
				}
			} catch (Exception e) {
				System.err.println(e.getMessage());
			}
		}

		final boolean isErrorStream;
		final BufferedReader reader;
	}

	public CommandLineTool(List<String> commands) throws Exception {
		ProcessBuilder processBuilder = new ProcessBuilder();
		processBuilder.command(commands);
		java.lang.Process process = processBuilder.start();
		final StreamGrabber outputGrabber = new StreamGrabber(process.getInputStream(), false);
		final StreamGrabber errorGraber = new StreamGrabber(process.getErrorStream(), true);
		outputGrabber.start();
		errorGraber.start();
		this.errorCode = process.waitFor();
		outputGrabber.join();
		errorGraber.join();
	}

	public CommandLineTool(List<String> commands, Map<String, String> environment) throws Exception {
		ProcessBuilder processBuilder = new ProcessBuilder();
		processBuilder.command(commands);
		processBuilder.environment().clear();
		processBuilder.environment().putAll(environment);
		java.lang.Process process = processBuilder.start();
		final StreamGrabber outputGrabber = new StreamGrabber(process.getInputStream(), false);
		final StreamGrabber errorGraber = new StreamGrabber(process.getErrorStream(), true);
		outputGrabber.start();
		errorGraber.start();
		this.errorCode = process.waitFor();
		outputGrabber.join();
		errorGraber.join();
	}

	public List<String> getOutput() {
		return output;
	}

	public List<String> getError() {
		return error;
	}

	public int getErrorCode() {
		return errorCode;
	}

	final List<String> output = new ArrayList<String>();
	final List<String> error = new ArrayList<String>();
	final int errorCode;
}
